#ifndef INC_MEMORYLEAKINDICATOR_H
#define INC_MEMORYLEAKINDICATOR_H
//-----------------------------------------------------------------------------
/*!
\file MemoryLeakIndicator.h
\brief
\if english
  Header for MemSpy::LeakMap
\endif

\if german
  Header für MemSpy::LeakMap
\endif
\author Helmut Jakoby
\copyright © 2021 Helmut Jakoby
<!-- @@TermsOfLicense BEGIN -->

<b>Terms of License</b>

This file is part of the MemoryLeakIndicator module.

MemoryLeakIndicator is free software:

This library is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.

<b>GNU Affero General Public License Usage</b><br>
You can redistribute it and/or modify it under the terms of the GNU Affero General
Public License as published by the Free Software Foundation; either version 3 of
the License, or (at your option) any later version. Please review the following
information to ensure the GNU Affero General Public License requirements will be
met: https://www.gnu.org/licenses/agpl-3.0.en.html .
<!-- @@TermsOfLicense END -->
*/
//-----------------------------------------------------------------------------
#ifdef _MSC_VER
  /* 'Bezeichner': Klasse 'Typ' benötigt eine DLL-Schnittstelle, die von... */
  #pragma warning( disable : 4251 )
#endif
//-----------------------------------------------------------------------------
#if __cplusplus >= 201103L
  #include <mutex>
#endif
#include <unordered_map>

#if defined(_MSC_VER) || defined(__MINGW32__) || defined(__MINGW64__)
  #include <basetsd.h>
#elif defined(__GNUC__)
  #define INT_PTR intptr_t
#endif
#include <stdarg.h>
#include <windows.h>
//-----------------------------------------------------------------------------
// DLL Define __MEM_LEAK_INDICATOR_EXPORT_DLL
#if defined(_MSC_VER) || defined(__MINGW32__) || defined(__MINGW64__)
  #ifdef __MEM_LEAK_INDICATOR_EXPORT_DLL
    #define __mem_leak_indicator_export_dll  __declspec(dllexport)
    #if defined(_MSC_VER)
      #pragma message ( "Benutze Export defines fuer __mem_leak_indicator_export_dll" )
    #endif
  #else
    #define __mem_leak_indicator_export_dll  //__declspec(dllimport)
    #if defined(_MSC_VER)
      #pragma message ( "Benutze Import defines fuer __mem_leak_indicator_export_dll" )
    #endif
  #endif
#elif defined(__GNUC__)
  #ifdef __MEM_LEAK_INDICATOR_EXPORT_DLL
    #define __mem_leak_indicator_export_dll  //__attribute__ ((visibility("default")))
  #else
    #define __mem_leak_indicator_export_dll  //__attribute__ ((visibility ("hidden")))
  #endif
#endif
//-----------------------------------------------------------------------------
/*!
\if english
  In Visual Studio extend the "Properties::C/C++::All options::Additional options:"
  with "/Zc:__cplusplus"!
\endif

\if german
  In Visual Studio die "Eigenschaften::C/C++::Alle Optionen::Zusätzliche Optionen:"
  um "/Zc:__cplusplus" erweitern!
\endif
*/
#ifndef NULL_PTR
  #if __cplusplus >= 201103L
    #define NULL_PTR nullptr
  #else
    #define NULL_PTR NULL
  #endif // __cplusplus >= 201103L
#endif
//-----------------------------------------------------------------------------
namespace MemSpy
{
  //---------------------------------------------------------------------------
  /*!
  \if english
    Type to tell the MemSpy::activateMemoryMonitoring (...) function whether an exception
    is thrown on error, or just a 0 as return value.
  \endif

  \if german
    Typ um der Funktion MemSpy::activateMemoryMonitoring (...) mitzuteilen, ob bei einem
    Fehler eine Exception geworfen wird, oder nur eine 0 als Rückgabewert.
  \endif
  */
  typedef bool TbExceptionMode;
  //---------------------------------------------------------------------------
  /*!
  \if english
    Function MemSpy::activateMemoryMonitoring (...) returns a 0 as return value in case
    of an error.
  \endif

  \if german
    Funktion MemSpy::activateMemoryMonitoring (...) liefert bei einem Fehler eine 0 als
    Rückgabewert.
  \endif
  */
  const TbExceptionMode NOTHROW = false;
  //---------------------------------------------------------------------------
  /*!
  \if english
    Function MemSpy::activateMemoryMonitoring (...) throws a std::range_error on error.
  \endif

  \if german
    Funktion MemSpy::activateMemoryMonitoring (...) wirft bei einem Fehler eine std::range_error.
  \endif
  */
  const TbExceptionMode THROW = true;
  //---------------------------------------------------------------------------

  //---------------------------------------------------------------------------
  /*!
  \if english
    Type to tell the MemSpy::finishMemoryMonitoring (...) function whether to output to
    the console.
  \endif

  \if german
    Typ um Funktion MemSpy::finishMemoryMonitoring (...) mitzuteilen, ob die Ausgabe in
    der Konsole ausgegeben wird.
  \endif
  */
  typedef int TiOutput;
  //---------------------------------------------------------------------------
  /*!
  \if english
    Function MemSpy::finishMemoryMonitoring (...) does not output anything. This can be
    useful if an output is to be printed to a GUI window.
  \endif

  \if german
    Funktion MemSpy::finishMemoryMonitoring (...) gibt nichts aus. Das kann sinnvoll sein,
    wenn eine Ausgabe in eine GUI-Fenster ausgegeben werden soll.
  \endif
  */
  const TiOutput OUTPUT_NON = 0;
  //---------------------------------------------------------------------------
  /*!
  \if english
    Function MemSpy::finishMemoryMonitoring (...) outputs its information to the console.
  \endif

  \if german
    Funktion MemSpy::finishMemoryMonitoring (...) gibt ihre Informationen in der Konsole
    aus.
  \endif
  */
  const TiOutput OUTPUT_CONSOLE = 1;
  //---------------------------------------------------------------------------
  /*!
  \if english
    Function MemSpy::finishMemoryMonitoring (...) outputs its information iin the output
    window.
  \endif

  \if german
    Funktion MemSpy::finishMemoryMonitoring (...) gibt ihre Informationen iim Ausgabefenster
    aus.
  \endif
  */
  const TiOutput OUTPUT_TRACE = 2;
  //---------------------------------------------------------------------------

  //---------------------------------------------------------------------------
  /*!
  \if english
    To iterate through LeakMap::m_ObjectList.
  \endif
    
  \if german
    Zum itereieren durch LeakMap::m_ObjectList.
  \endif
  */
  enum SeekMode
  {
    /*!
    \if english
      The first element is meant.
    \endif
      
    \if german
      Das erste Element ist gemeint.
    \endif
    */
    START,
    /*!
    \if english
      The next element after the current element is meant.
    \endif
      
    \if german
      Das nächste Element nach dem aktuellen Element ist gemeint.
    \endif
    */
    NEXT,
    /*!
    \if english
      The current element is meant.
    \endif
      
    \if german
      Das aktuelle Element ist gemeint.
    \endif
    */
    CURRENT
  };
  //---------------------------------------------------------------------------

  //---------------------------------------------------------------------------
  /*!
  \brief
  \if english
    Class to be able to write a TRACE output to the output window also in Releas
    mode. <br>
    Found in 
  \endif
  \if german
    Klasse, um eine TRACE-Ausgaben auch im Releas-Modus ins Ausgabefenster
    schreiben zu können. <br>
    Gefunden in 
  \endif
  http://blog.m-ri.de/index.php/2009/08/09/vs-tippstricks-einfache-ausgabe-mit-trace-auch-in-der-release-version/
  
  \author Martin Richter
  */
  class __mem_leak_indicator_export_dll Trace
  {
    public:
      void operator()( PCSTR pszFormat, ... );
      void operator()( PCWSTR pszFormat, ... );
  };

  //---------------------------------------------------------------------------
  /*!
  \brief
  \if english
    Class to access forgotten objects in memory. <br>
    Inspired by Rainer Grimm and Mouaz Chamieh.
  \endif

  \if german
    Klasse um vergessene Objekte im Speicher im Zugriff zu haben. <br>
    Inspiriert von Rainer Grimm und Mouaz Chamieh.
  \endif

  \sa https://www.grimm-jaud.de/index.php/blog/operator-new-und-delete-ueberladen
    und https://github.com/muazsh/MemoryLeakManager
  
  \author Helmut Jakoby
  \copyright © 2021 Helmut Jakoby
  */
  class __mem_leak_indicator_export_dll LeakMap
  {
    public:
      //-----------------------------------------------------------------------
      // >>>>>>>>>>>>>>>>>>>>>>>>> MemoryItem >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
      //-----------------------------------------------------------------------
      /*! \brief
      \if english
        Reference to the instantiated object with file name and line where it
        was instantiated.
      \endif

      \if german
        Referenz auf das instantiierte Objekt mit Dateinamen und Zeile, in der
        es instantiiert wurde.
      \endif

      \copyright © 2021 Helmut Jakoby
      \author Helmut Jakoby
      */
      struct MemoryItem
      {
        //============== Attribute
        //---------------------------------------------------------------------
        /*!
        \if english
          Address of the instantiated object cast to INT_PTR.
        \endif

        \if german
          Auf INT_PTR gecastete Adresse des instantiierten Objekts.
        \endif
        */
        INT_PTR m_pItemPtr;
        //---------------------------------------------------------------------
        /*!
        \if english
          The name of the file in which the object was instantiated.
        \endif

        \if german
          Der Name der Datei, in der das Objekt instantiiert wurde.
        \endif
        */
        char const * m_FileName;
        //---------------------------------------------------------------------
        /*!
        \if english
          The line number in which the object was instantiated.
        \endif

        \if german
          Die Zeilennummer, in der das Objekt instantiiert wurde.
        \endif
        */
        int m_iLineNumber;
        //---------------------------------------------------------------------
      };
      //-----------------------------------------------------------------------
      // <<<<<<<<<<<<<<<<<<<<<<<<< MemoryItem <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
      //-----------------------------------------------------------------------

    private:
      //============== Attribute
      //-----------------------------------------------------------------------
      /*!
      \if english
        To have thread-safe access to the list of objects, access is synchronized
        via this local mutex.
      \endif
    
      \if german
        Um einen Thread-sicheren Zugriff auf die Liste der Objekte zu haben, wird
        über diesen lokalen Mutex der Zugriff synchronisiert.
      \endif
      */
      #if __cplusplus >= 201103L
        std::mutex m_LocalCriticalSection;
      #else
        CRITICAL_SECTION m_LocalCriticalSection;
      #endif
      //-----------------------------------------------------------------------
      /*!
      \if english
        The list of MemoryItem objects. The index is the address, cast to INT_PTR,
        of the individual objects.
      \endif
    
      \if german
        Die Liste der MemoryItem-Objekte. Der Index ist die, auf INT_PTR gecastete
        Adresse der einzelnen Objekte.
      \endif
      */
      std::unordered_map< INT_PTR, MemoryItem * > m_ObjectList;
      //-----------------------------------------------------------------------
      /*!
      \if english
        An internal iterator to the MemoryItem object list.
      \endif
    
      \if german
        Ein interner Iterator auf die Liste der MemoryItem-Objekte.
      \endif
      */
      std::unordered_map< INT_PTR, MemoryItem * >::iterator m_ObjectListIterator;
      //-----------------------------------------------------------------------
      /*!
      \if english
        A counter that is incremented for each inserted object.
      \endif
    
      \if german
        Ein Zähler, der bei jedem eingefügten Objekt inkrementiert wird.
      \endif
      */
      unsigned long long m_ullInsertCounter;
      //-----------------------------------------------------------------------

      //========== static Attribute =========
      //-----------------------------------------------------------------------
      /*!
      \if english
        Only if true, LeakMap is used in new (...) and delete (...). Is necessary
        to not recursively process own calls of new and delete here.
      \endif
    
      \if german
        Nur wenn true, wird LeakMap in new (...) und delete (...) genutzt. Ist
        notwendig, um eigene Aufrufe von new und delete nicht rekursiv hier zu
        bearbeiten.
      \endif
      */
       static bool ms_bUseLeakMap;
      //-----------------------------------------------------------------------
      /*!
      \if english
        The reference counter to the only instance of this class.
      \endif
    
      \if german
        Der Referenzzähler auf die einzige Instanz dieser Klasse.
      \endif
      */
      static int ms_iReferenceCounter;
      //-----------------------------------------------------------------------

    protected:
      //============== Konstruktoren
      //-----------------------------------------------------------------------
      /*!
      \if english
        The standard constructor is not available. You can get an instance via
        LeakMap::create().
      \endif
    
      \if german
        Der Standard-Konstruktor steht nicht zur Verfügung. Eine Instanz bekommt
        man über LeakMap::create().
      \endif
      */
      LeakMap();
      //-----------------------------------------------------------------------
      /*!
      \if english
        The destructor is not available. An LeakMap object may only be
        removed from memory with LeakMap::destroy().
      \endif
    
      \if german
        Der Destruktor steht nicht zur Verfügung. Ein LeakMap-Objekt darf
        nur über LeakMap::destroy() aus dem Speicher entfernt werden.
      \endif
      */
      virtual ~LeakMap();
      //-----------------------------------------------------------------------

    private:
      //============== Konstruktoren
      //-----------------------------------------------------------------------
      /*!
      \if english
        The copy constructor is not available.
      \endif

      \if german
        Der Copy-Konstruktor steht nicht zur Verfügung.
      \endif
      */
      LeakMap( const LeakMap & );
      //-----------------------------------------------------------------------

    public:
      //========== static Methoden =========
      //-----------------------------------------------------------------------
      /*!
      \if english
        Instantiation function. Since LeakMap is a singleton class, this method
        instantiates it once and then only returns the pointer to the instance.
        \return The only instance of this class.
        \attention The delivered instance must be removed from memory with
          LeakMap::destroy().
      \endif
    
      \if german
        Instanziierungsfunktion. Da es sich bei LeakMap um eine Singleton-Class
        handelt, wird diese von dieser Methode einmalig instanziiert und dann
        nur noch der Pointer auf die Instanz geliefert.
        \return Die einzige Instanz dieser Klasse.
        \attention Die gelieferte Instanz muss mit LeakMap::destroy() aus
          dem Speicher entfernt werden.
      \endif
      */
      static LeakMap * create();
      //-----------------------------------------------------------------------
      /*!
      \if english
        If the calling instance is the last one with a reference to the only
        instance of LeakMap, this instance is removed from memory.
        \return Return is the number of still-references to this instance.
      \endif
    
      \if german
        Wenn die aufrufende Instanz die letzte mit einer Referenz auf die einzige
        Instanz von LeakMap ist, wird diese Instanz aus dem Speicher entfernt.
        \return Rückgabe ist die Anzahl der Noch-Referenzen auf diese Instanz.
      \endif
      */
      static int destroy();
      //-----------------------------------------------------------------------
      /*!
      \if english
        It removes the only instance of LeakMap from memory without querying
        #ms_iReferenceCounter.
        \attention This should only be done at the end of the main.
      \endif
    
      \if german
        Es wird ohne Abfrage von #ms_iReferenceCounter die einzige Instanz von
        LeakMap aus dem Speicher entfernt.
        \attention Das sollte nur am Ende der main erfolgen.
      \endif
      */
      static void destroyAbsolutely();
      //-----------------------------------------------------------------------
      /*!
      \if english
        Returns #g_pVoidLeakMap cast to LeakMap if instantiated.
      \endif
    
      \if german
        Liefert #g_pVoidLeakMap gecastet auf LeakMap, wenn instantiiert.
      \endif
      */
      static LeakMap * getGlobalLeakMapInstance();
      //-----------------------------------------------------------------------

    public:
      //============== Methoden
      //-----------------------------------------------------------------------
      /*!
      \if english
        Called in MemSpy::finishMemoryMonitoring (...) to disable memory monitoring.
        \param[in] bAllowed If true, memory monitoring is enabled, otherwise disabled. 
      \endif
    
      \if german
        Wird in MemSpy::finishMemoryMonitoring (...) aufgerufen, um die Speicherüberwachung
        zu deaktivieren.
        \param[in] bAllowed Wenn true, wird die Speicherüberwachung aktiviert,
          ansonsten deaktiviert.
      \endif
      \sa #ms_bUseLeakMap
      */
      void setLeakMapUseAllowed( bool bAllowed = true );
      //-----------------------------------------------------------------------
      /*!
      \if english
        Called in operator new( size_t, const char *, int ) or operator delete( void * )
        to determine if LeakMap should be used.
        \return If true, the newly instantiated or removed from memory object is
          processed by LeakMap.
      \endif
    
      \if german
        Wird in operator new( size_t, const char *, int ) oder
        operator delete( void * ) aufgerufen, um zu bestimmen ob LeakMap
        genutzt werden soll.
        \return Wenn true, wird das neu instantiierte bzw. aus dem Speicher entfernte
          Objekt von LeakMap verarbeitet.
      \endif
      \sa #ms_bUseLeakMap
      */
      bool leakMapUseAllowed() const;
      //-----------------------------------------------------------------------
      /*!
      \if english
        Returns the number of objects processed so far.
        \return See above.
      \endif
    
      \if german
        Liefert die Anzahl der bislang verarbeiteten Objekte.
        \return s.o.
      \endif
      */
      unsigned long long numberOfProcessedObjects() const;
      //-----------------------------------------------------------------------
      /*!
      \if english
        Retrurns the number of entries in #m_ObjectList.
        \return See above.
      \endif
    
      \if german
        Liefert die Anzahl der Einträge in #m_ObjectList.
        \return s.o.
      \endif
      */
      std::size_t sizeOfObjectList() const;
      //-----------------------------------------------------------------------
      /*!
      \if english
        Returns a MemoryItem object from the list of MemoryItem objects related
        to eMode.
        \param[in] eMode The Seekmode

        <B>Parameter examples:</B>
        \li #START    = get first element
        \li #CURRENT  = returns current element
        \li #NEXT     = returns next element

        \return If found, the MemoryItem object, otherwise \ref NULL_PTR "NULL_PTR".
        \attention The calling instance must not remove the supplied object from
          memory.
      \endif
    
      \if german
        Liefert ein MemoryItem-Objekt aus der Liste der MemoryItem-Objekte in
        Bezug auf eMode.
        \param[in] eMode Der Suchmodus

        <B>Parameterbeispiele:</B>
        \li #START     = liefert erstes Element
        \li #CURRENT   = liefert aktuelles Element
        \li #NEXT      = liefert nächstes Element

        \return Wenn gefunden das MemoryItem-Objekt, ansonsten \ref NULL_PTR "NULL_PTR".
        \attention Die aufrufende Instanz darf das gelieferte Objekt nicht aus
          dem Speicher entfernen.
      \endif
      */
      MemoryItem * getSpyObject( SeekMode eMode );
      //-----------------------------------------------------------------------
      /*!
      \if english
        Returns a MemoryItem object from the list of MemoryItem objects which
        references the object with the passed address.
        \param[in] ipAddress The address of the searched object cast on INT_PTR.
        \return If found the MemoryItem object, otherwise \ref NULL_PTR "NULL_PTR".
        \attention The calling instance must not remove the supplied object from
          memory.
      \endif
    
      \if german
        Liefert ein MemoryItem-Objekt aus der Liste der MemoryItem-Objekte welches
        das Objekt mit der übergebenen Adresse referenziert.
        \param[in] ipAddress Die auf INT_PTR gecastete Adresse des gesuchten
          Objektes.
        \return Wenn gefunden das MemoryItem-Objekt, ansonsten \ref NULL_PTR "NULL_PTR".
        \attention Die aufrufende Instanz darf das gelieferte Objekt nicht aus dem
          Speicher entfernen.
      \endif
      */
      MemoryItem * getSpyObject( INT_PTR ipAddress );
      //-----------------------------------------------------------------------
      /*!
      \if english
        Inserts a MemoryItem object into the list of MemoryItem objects.

        \param[in] pNewObject The pointer to the new object to be inserted.
        \param[in] chFileName The name of the file in which the new object was created.
        \param[in] iLine The line number in which the new object was created.
      \endif
    
      \if german
        Fügt ein MemoryItem-Objekt in die Liste der MemoryItem-Objekte ein.

        \param[in] pNewObject Der Zeiger auf das neue einzufügende Objekt.
        \param[in] chFileName Der Name der Datei, in dem das neue Objekt erzeugt wurde.
        \param[in] iLine Die Zeilennummer, in dem das neue Objekt erzeugt wurde.
      \endif
      */
      void insertSpyObject( void * pNewObject, char const * chFileName, int iLine );
      //-----------------------------------------------------------------------
      /*!
      \if english
        Removes a MemoryItem object from the list of MemoryItem objects.
        \param[in] pObjectToDelete Reference to the object to delete.
      \endif
    
      \if german
        Entfernt ein MemoryItem-Objekt aus der Liste der MemoryItem-Objekte.
        \param[in] pObjectToDelete Referenz auf das zu löschende Objekt.
      \endif
      */
      void removeSpyObject( void * pObjectToDelete );
      //-----------------------------------------------------------------------

    private:
      //============== Operatoren
      //-----------------------------------------------------------------------
      /*!
      \if english
        The assignment operator is not available.
      \endif

      \if german
        Der Zuweisungsoperator steht nicht zur Verfügung.
      \endif
      */
      LeakMap & operator= ( const LeakMap & );
      //-----------------------------------------------------------------------
  };
  //---------------------------------------------------------------------------
  
  //---------------------------------------------------------------------------
  /*!
  \if english
    Enables storage monitoring and instantiates the LeakMap. <br>
    This function should be called in the main at the beginning before
    instantiating objects.
    \param[in] bExceptionMode See MemSpy::NoThrow and MemSpy::Throw.
    \return If bExceptionMode == MemSpy::NOTHROW, a NULL_PTR is returned on error,
      otherwise a pointer to the MemSpy::LeakMap. 
    \exception If this function is called again, a std::range_error is thrown
      if bExceptionMode == MemSpy::THROW.
    \attention If a pointer to the MemSpy::LeakMap is supplied, the MemSpy::LeakMap
      must not be removed from memory by the calling instance. <br>
     This may only be initiated with MemSpy::finishMemoryMonitoring (...),
      MemSpy::LeakMap::destroy() or MemSpy::LeakMap::destroyAbsolutely(). 
  \endif
    
  \if german
    Aktiviert die Speicherüberwachung und instantiiert die LeakMap. <br>
    Diese Funktion sollte in der main am Anfang vor dem Instantiieren von
    Objekten aufgerufen werden.
    \param[in] bExceptionMode Siehe MemSpy::NoThrow und MemSpy::Throw.
    \return Wenn bExceptionMode == MemSpy::NOTHROW, wird bei einem Fehler ein
      NULL_PTR zurückgeliefert, ansonsten ein Zeiger auf die MemSpy::LeakMap.
    \exception Wenn diese Funktion nochmals aufgerufen wird, wird bei
      bExceptionMode == MemSpy::THROW ein std::range_error geworfen.
    \attention Wenn ein Zeiger auf die MemSpy::LeakMap geliefert wird, darf die
      MemSpy::LeakMap nicht von der aufrufenden Instanz aus dem Speicher entfernt
      werden. <br>
      Dieses darf nur mit MemSpy::finishMemoryMonitoring (...),
      MemSpy::LeakMap::destroy() oder MemSpy::LeakMap::destroyAbsolutely()
      initiiert werden.
  \endif
  */
  __mem_leak_indicator_export_dll MemSpy::LeakMap * activateMemoryMonitoring( MemSpy::TbExceptionMode bExceptionMode = MemSpy::THROW );
  //---------------------------------------------------------------------------
  /*!
  \if english
    Stops memory monitoring and removes the LeakMap from memory. <br>
    This function should be called in main at the end after object instantiation,
    so that information about memory leaks can be output if necessary.
    \param[in] iOutput See MemSpy::OUTPUT_NON and MemSpy::OUTPUT_CONSOLE.
  \endif
    
  \if german
    Beendet die Speicherüberwachung und entfernt die LeakMap aus dem Speicher. <br>
    Diese Funktion sollte in der main am Ende nach dem Instantiieren von Objekten
    aufgerufen werden, damit ggf. Informationen über Speicherlecks ausgegeben
    werden können.
    \param[in] iOutput Siehe MemSpy::OUTPUT_NON und MemSpy::OUTPUT_CONSOLE. Parameter können Oderverknüpft werden.
  \endif
  */
  void __mem_leak_indicator_export_dll finishMemoryMonitoring( MemSpy::TiOutput iOutput = MemSpy::OUTPUT_CONSOLE );
  //---------------------------------------------------------------------------
} // namespace MemSpy
//-----------------------------------------------------------------------------
#ifdef _MSC_VER
  /* 'Bezeichner': Klasse 'Typ' benötigt eine DLL-Schnittstelle, die von... */
  #pragma warning( default : 4251 )
#endif
//-----------------------------------------------------------------------------
#endif
