//-----------------------------------------------------------------------------
/*!
\file MemoryLeakIndicator.cpp
\brief Source for MemSpy::LeakMap
\author Helmut Jakoby
\copyright © 2021 Helmut Jakoby
<!-- @@TermsOfLicense BEGIN -->

<b>Terms of License</b>

This file is part of the MemoryLeakIndicator module.

MemoryLeakIndicator is free software:

This library is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.

<b>GNU Affero General Public License Usage</b><br>
You can redistribute it and/or modify it under the terms of the GNU Affero General Public
License as published by the Free Software Foundation; either version 3 of the
License, or (at your option) any later version. Please review the following
information to ensure the GNU Affero General Public License requirements will be
met: https://www.gnu.org/licenses/agpl-3.0.en.html .
<!-- @@TermsOfLicense END -->
*/
//-----------------------------------------------------------------------------
#include "MemoryLeakIndicator.h"
#include "MemoryLeakIndicatorGlobals.h"
#include <iostream>
#include <sstream>
//-----------------------------------------------------------------------------
namespace MemSpy
{
  //---------------------------------------------------------------------------
  void Trace::operator()( PCSTR pszFormat, ... )
  {
    va_list ptr;
    va_start( ptr, pszFormat );
    //TraceV( pszFormat,ptr );
    char szBuffer[1024];
    _vsnprintf( szBuffer, _countof( szBuffer ), pszFormat, ptr );
    OutputDebugStringA( szBuffer );
    va_end( ptr );
  }
  //---------------------------------------------------------------------------
  void Trace::operator()( PCWSTR pszFormat, ... )
  {
    va_list ptr;
    va_start( ptr, pszFormat );
    //TraceV( pszFormat,ptr );
    wchar_t szBuffer[1024];
    _vsnwprintf( szBuffer, _countof( szBuffer ), pszFormat, ptr );
    OutputDebugStringW( szBuffer );
    va_end( ptr );
  }
  //---------------------------------------------------------------------------

  //---------------------------------------------------------------------------
  int LeakMap::ms_iReferenceCounter = 0;
  //---------------------------------------------------------------------------
  bool LeakMap::ms_bUseLeakMap = true;
  //---------------------------------------------------------------------------
  LeakMap::LeakMap()
  : m_ullInsertCounter( 0 )
  {
    #if __cplusplus < 201103L
      InitializeCriticalSection( & m_LocalCriticalSection );
    #endif
    m_ObjectListIterator = m_ObjectList.end();
  }
  //---------------------------------------------------------------------------
  LeakMap::~LeakMap()
  {
    #if __cplusplus >= 201103L
      m_LocalCriticalSection.lock();
    #else
      EnterCriticalSection( & m_LocalCriticalSection );
    #endif
    for ( std::unordered_map< INT_PTR, MemoryItem * >::iterator t_ObjectListIterator = m_ObjectList.begin();
          t_ObjectListIterator != m_ObjectList.end();
          ++ t_ObjectListIterator )
    {
      free( (MemoryItem *) t_ObjectListIterator->second );
    }
    m_ObjectList.clear();
    #if __cplusplus >= 201103L
      m_LocalCriticalSection.unlock();
    #else
      LeaveCriticalSection( & m_LocalCriticalSection );
      DeleteCriticalSection( & m_LocalCriticalSection );
    #endif
  }
  //---------------------------------------------------------------------------
  LeakMap * LeakMap::create()
  {
    // wenn noch keine Instanz, dann neu...
    if ( g_pVoidLeakMap == NULL_PTR )
    {
      ms_bUseLeakMap = false;
      // erste und einzige Instanz von Klasse instanziieren; wirft evtl. Exception!
      g_pVoidLeakMap = new MemSpy::LeakMap();
      ms_bUseLeakMap = true;
    }
    ms_iReferenceCounter++;  // ReferenzZaehler hochzählen
    return static_cast<LeakMap*>( g_pVoidLeakMap );
  }
  //---------------------------------------------------------------------------
  int LeakMap::destroy()
  {
    // Abfrage auf > 0, damit bei einem versehendlichem Destruieren zuviel nichts passiert
    if ( ms_iReferenceCounter > 0 ) ms_iReferenceCounter--;  // RefernzZaehler runterzählen
    // wenn letzter instanziierer löschen möchte...
    if ( ms_iReferenceCounter == 0 )
    {
      ms_bUseLeakMap = false;
      LeakMap * t_pDeleteLeakMap = static_cast<LeakMap*>( g_pVoidLeakMap );
      if ( t_pDeleteLeakMap )
      {
        delete( t_pDeleteLeakMap );  // Instanz aus Speicher entfernen
      }
      // und g_pVoidLeakMap unbedingt auf 0 setzen, damit nächste instanziierung klappt.
      g_pVoidLeakMap = NULL_PTR;
    }
    return ms_iReferenceCounter;
  }
  //---------------------------------------------------------------------------
  void LeakMap::destroyAbsolutely()
  {
    ms_bUseLeakMap = false;
    LeakMap * t_pDeleteLeakMap = static_cast<LeakMap*>( g_pVoidLeakMap );
    if ( t_pDeleteLeakMap )
    {
      delete( t_pDeleteLeakMap );  // Instanz aus Speicher entfernen
    }
    // und g_pVoidLeakMap unbedingt auf 0 setzen, damit nächste instanziierung klappt.
    g_pVoidLeakMap = NULL_PTR;
  }
  //---------------------------------------------------------------------------
  LeakMap * LeakMap::getGlobalLeakMapInstance()
  {
    if ( g_pVoidLeakMap )
    {
      return static_cast<LeakMap*>( g_pVoidLeakMap );
    }
    return NULL_PTR;
  }
  //---------------------------------------------------------------------------
  void LeakMap::setLeakMapUseAllowed( bool bAllowed )
  {
    #if __cplusplus >= 201103L
      m_LocalCriticalSection.lock();
    #else
      EnterCriticalSection( & m_LocalCriticalSection );
    #endif
    ms_bUseLeakMap = bAllowed;
    #if __cplusplus >= 201103L
      m_LocalCriticalSection.unlock();
    #else
      LeaveCriticalSection( & m_LocalCriticalSection );
    #endif
  }
  //---------------------------------------------------------------------------
  bool LeakMap::leakMapUseAllowed() const
  {
    return ms_bUseLeakMap;
  }
  //---------------------------------------------------------------------------
  unsigned long long LeakMap::numberOfProcessedObjects() const
  {
    return m_ullInsertCounter;
  }
  //---------------------------------------------------------------------------
  std::size_t LeakMap::sizeOfObjectList() const
  {
    return m_ObjectList.size();
  }
  //---------------------------------------------------------------------------
  #ifdef _MSC_VER
    // Der Enumerationstyp "xxx" umfasst keine Bereichseinschränkung. Geben Sie "enum class" Vorrang vor "enum" (Enum.3)
    #pragma warning( disable : 26812 )
  #endif
  //---------------------------------------------------------------------------
  LeakMap::MemoryItem * LeakMap::getSpyObject( SeekMode eMode )
  {
    #if __cplusplus >= 201103L
      m_LocalCriticalSection.lock();
    #else
      EnterCriticalSection( & m_LocalCriticalSection );
    #endif

    MemoryItem * t_pRetVal = NULL_PTR;

    if ( eMode == START )
    {
      m_ObjectListIterator = m_ObjectList.begin();
    }
    else if ( eMode == NEXT )
    {
      m_ObjectListIterator++;
    }
    if ( m_ObjectListIterator != m_ObjectList.end() )
    {
      t_pRetVal = (*m_ObjectListIterator).second;
    }
    #if __cplusplus >= 201103L
      m_LocalCriticalSection.unlock();
    #else
      LeaveCriticalSection( & m_LocalCriticalSection );
    #endif
    return t_pRetVal;
  }
  //---------------------------------------------------------------------------
  #ifdef _MSC_VER
    // Der Enumerationstyp "xxx" umfasst keine Bereichseinschränkung. Geben Sie "enum class" Vorrang vor "enum" (Enum.3)
    #pragma warning( default : 26812 )
  #endif
  //---------------------------------------------------------------------------
  LeakMap::MemoryItem * LeakMap::getSpyObject( INT_PTR ipAddress )
  {
    #if __cplusplus >= 201103L
      m_LocalCriticalSection.lock();
    #else
      EnterCriticalSection( & m_LocalCriticalSection );
    #endif

    MemoryItem  * t_pRetVal = NULL_PTR;

    std::unordered_map< INT_PTR, MemoryItem * >::iterator   t_Iterator = m_ObjectList.find( ipAddress );

    if ( t_Iterator != m_ObjectList.end() )
    {
      t_pRetVal = (*t_Iterator).second;
    }
    #if __cplusplus >= 201103L
      m_LocalCriticalSection.unlock();
    #else
      LeaveCriticalSection( & m_LocalCriticalSection );
    #endif
    return t_pRetVal;
  }
  //---------------------------------------------------------------------------
  void LeakMap::insertSpyObject( void * pNewObject, char const * chFileName, int iLine )
  {
    #if __cplusplus >= 201103L
      m_LocalCriticalSection.lock();
    #else
      EnterCriticalSection( & m_LocalCriticalSection );
    #endif

    MemoryItem  * t_pNewMemoryItem = ( MemoryItem * ) std::malloc( sizeof( MemoryItem ) );

    if ( t_pNewMemoryItem )
    {
      ++m_ullInsertCounter;
      t_pNewMemoryItem->m_pItemPtr = (INT_PTR) pNewObject;
      t_pNewMemoryItem->m_FileName = chFileName;
      t_pNewMemoryItem->m_iLineNumber = iLine;

      if ( m_ObjectList.find( (INT_PTR) t_pNewMemoryItem->m_pItemPtr ) == m_ObjectList.end() )
      {
        // m_ObjectListIterator kann auf NULL_PRT zeigen, wenn Liste erweitert wird...
        MemoryItem  * t_pTempRef = NULL_PTR;

        if ( m_ObjectListIterator != m_ObjectList.end() )
        {
          t_pTempRef = (*m_ObjectListIterator).second;
        }
        ms_bUseLeakMap = false;
        m_ObjectList.insert( std::pair<INT_PTR, MemoryItem *>( (INT_PTR) t_pNewMemoryItem->m_pItemPtr, const_cast<MemoryItem*>( t_pNewMemoryItem ) ) );
        ms_bUseLeakMap = true;
        if ( t_pTempRef )
        {
          m_ObjectListIterator = m_ObjectList.find( (INT_PTR) t_pTempRef );
        }
      }
      else
      {
        free( t_pNewMemoryItem );
      }
    }
    #if __cplusplus >= 201103L
      m_LocalCriticalSection.unlock();
    #else
      LeaveCriticalSection( & m_LocalCriticalSection );
    #endif
  }
  //---------------------------------------------------------------------------
  void LeakMap::removeSpyObject( void * pObjectToDelete )
  {
    #if __cplusplus >= 201103L
      m_LocalCriticalSection.lock();
    #else
      EnterCriticalSection( & m_LocalCriticalSection );
    #endif
    std::unordered_map< INT_PTR, MemoryItem * >::iterator   t_ObjectListIterator = m_ObjectList.find( (INT_PTR) pObjectToDelete );

    if ( t_ObjectListIterator != m_ObjectList.end() )
    {
      // wenn m_ObjectListIterator auf den zu löschenden zeigt, verschieben damit nicht Zugriff auf NULL_PTR
      if ( m_ObjectListIterator != m_ObjectList.end() && ( m_ObjectListIterator->second == t_ObjectListIterator->second ) )
      {
        ++ m_ObjectListIterator;
      }
      free( (MemoryItem *) t_ObjectListIterator->second );
      ms_bUseLeakMap = false;
      m_ObjectList.erase( t_ObjectListIterator );
      ms_bUseLeakMap = true;
    }
    #if __cplusplus >= 201103L
      m_LocalCriticalSection.unlock();
    #else
      LeaveCriticalSection( & m_LocalCriticalSection );
    #endif
  }
  //---------------------------------------------------------------------------
  MemSpy::LeakMap * activateMemoryMonitoring( MemSpy::TbExceptionMode bExceptionMode )
  {
    MemSpy::LeakMap * t_pRetVal = NULL_PTR;

    if ( g_pVoidLeakMap )
    {
      if ( bExceptionMode == MemSpy::THROW )
      {
        throw std::range_error("Too many activations");
      }
      else
      {
        return t_pRetVal;
      }
    }
    try
    {
      t_pRetVal = LeakMap::create();
    }
    catch ( std::bad_alloc & )
    {
      t_pRetVal = NULL_PTR;
      if ( bExceptionMode == MemSpy::NOTHROW )
      {
        return t_pRetVal;
      }
    }
    return t_pRetVal;
  }
  //---------------------------------------------------------------------------
  void finishMemoryMonitoring( MemSpy::TiOutput iOutput )
  {
    if ( g_pVoidLeakMap )
    {
      MemSpy::LeakMap * t_pLeakMap = static_cast<LeakMap*>( g_pVoidLeakMap );
      std::string       t_sStars       ( "*********************************************************************\n" ),
                        t_sNoMemLeaks  ( "* No memory leaks detected!                                         *\n" ),
                        t_sMemLeaks    ( "* Memory leaks detected!                                            *\n" ),
                        t_sMemLeakLine1( "* Memory leak in " ),
                        t_sMemLeakLine2( ", file " ),
                        t_sMemLeakLine3( ", line " );

      t_pLeakMap->setLeakMapUseAllowed( false );
      if ( iOutput & OUTPUT_CONSOLE )
      {
        std::cout << t_sStars;         // "*********************************************************************\n"
        // Display memory leaks through MemSpy::LeakMap
        if ( t_pLeakMap->sizeOfObjectList() )
        {
          MemSpy::LeakMap::MemoryItem  * t_pMemoryItem = t_pLeakMap->getSpyObject( MemSpy::START );
          if ( t_pMemoryItem )
          {
            std::cout << t_sMemLeaks;  // "* Memory leaks detected!                                            *\n"
            std::cout << t_sStars;     // "*********************************************************************\n"
            do
            {
              std::cout << t_sMemLeakLine1 << t_pMemoryItem->m_pItemPtr << t_sMemLeakLine2 << t_pMemoryItem->m_FileName << t_sMemLeakLine3 <<t_pMemoryItem->m_iLineNumber << "\n";
              t_pMemoryItem = t_pLeakMap->getSpyObject( MemSpy::NEXT );
            } while ( t_pMemoryItem );
            std::cout << t_sStars;     // "*********************************************************************\n"
          }
        }
        else
        {
          std::cout << t_sNoMemLeaks;  // "* No memory leaks detected!                                         *\n"
          std::cout << t_sStars;       // "*********************************************************************\n"
        }
      }
      if ( iOutput & OUTPUT_TRACE )
      {
        Trace TRACE;
        TRACE( t_sStars.c_str() );         // "*********************************************************************\n"
        // Display memory leaks through MemSpy::LeakMap
        if ( t_pLeakMap->sizeOfObjectList() )
        {
          MemSpy::LeakMap::MemoryItem  * t_pMemoryItem = t_pLeakMap->getSpyObject( MemSpy::START );
          if ( t_pMemoryItem )
          {
            TRACE( t_sMemLeaks.c_str() );  // "* Memory leaks detected!                                            *\n"
            TRACE( t_sStars.c_str() );     // "*********************************************************************\n"
            do
            {
              std::ostringstream  t_ossOutput;

              t_ossOutput << t_sMemLeakLine1 << t_pMemoryItem->m_pItemPtr << t_sMemLeakLine2 << t_pMemoryItem->m_FileName << t_sMemLeakLine3 <<t_pMemoryItem->m_iLineNumber << "\n";
              TRACE( t_ossOutput.str().c_str() );
              t_pMemoryItem = t_pLeakMap->getSpyObject( MemSpy::NEXT );
            } while ( t_pMemoryItem );
            TRACE( t_sStars.c_str() );     // "*********************************************************************\n"
          }
        }
        else
        {
          TRACE( t_sNoMemLeaks.c_str() );  // "* No memory leaks detected!                                         *\n"
          TRACE( t_sStars.c_str() );       // "*********************************************************************\n"
        }
      }
    }
    MemSpy::LeakMap::destroyAbsolutely();
  }
  //---------------------------------------------------------------------------
} // namespace MemSpy

